//
//  WindowSizing.m
//
//  Created by Cameron Hayne on Sat Nov 15 2003.
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "WindowSizing.h"

// category on NSWindow to supply resizing methods
@implementation NSWindow (WindowSizing)

- (float)toolbarHeight
{
    // (derived from code on Apple's web site)
    float toolbarHeight = 0.0;
    NSToolbar *toolbar = [self toolbar];

    if (toolbar && [toolbar isVisible])
    {
        NSRect contentRect = [self contentRectForFrameRect:[self frame]];
        toolbarHeight = NSHeight(contentRect)
                      - NSHeight([[self contentView] frame]);
    }

    return toolbarHeight;
}

// maximum size this window can be without overlapping the menubar or the dock
- (NSSize)maxUsableWindowSize
{
    return [[self screen] visibleFrame].size;
}

// maximum size the given view can get to when this window is at its max size
- (NSSize)maxUsableViewSize:(NSView *)aView
{
    NSSize winSize  = [self frame].size;
    NSSize maxWinSize = [self maxUsableWindowSize];
    float deltaWidth = maxWinSize.width - winSize.width;
    float deltaHeight = maxWinSize.height - winSize.height;
    NSSize viewSize = [aView frame].size;
    return NSMakeSize(viewSize.width + deltaWidth, viewSize.height + deltaHeight);
}

- (void)resizeForContentSize:(NSSize)desiredContentSize animate:(BOOL)flag
{
    NSRect contentRect = [self contentRectForFrameRect:[self frame]];

    // need to shift origin (bottomleft) so window stays in same place
    contentRect.origin.y += (contentRect.size.height - desiredContentSize.height);
    contentRect.size = desiredContentSize;

    NSRect frame = [self frameRectForContentRect:contentRect];
    [self setFrame: frame
           display: YES
           animate: flag];
}

- (void)changeWidthBy:(float)deltaWidth heightBy:(float)deltaHeight animate:(BOOL)flag
{
    NSRect contentRect = [self contentRectForFrameRect:[self frame]];
    
    // need to shift origin (bottomleft) so window stays in same place
    contentRect.origin.y -= deltaHeight;
    contentRect.size.width  += deltaWidth;
    contentRect.size.height += deltaHeight;
    
    NSRect frame = [self frameRectForContentRect: contentRect];
    [self setFrame: frame
           display: YES
           animate: flag];
}

- (void)freezeAspectRatio
{
    NSRect contentRect = [self contentRectForFrameRect:[self frame]];
    [self setContentAspectRatio: contentRect.size];
}

static int viewSortByPosition(id view1, id view2, void *context)

{
    float y1 = [view1 frame].origin.y;
    float y2 = [view2 frame].origin.y;
    
    if (y1 < y2)
    {
        return NSOrderedDescending;
    }
    else if (y1 > y2)
    {
        return NSOrderedAscending;
    }
    else
    {
        return NSOrderedSame;
    }
}

// Note: this assumes that the view being toggled is the full width of the window
// I.e. that the resizing will be in the vertical direction only
- (void) toggleViewAndResize: (NSView *)aView
{
    NSArray *subviews = [[self contentView] subviews];
    subviews = [subviews sortedArrayUsingFunction: viewSortByPosition
                                          context: nil];
    
    int numSubviews = [subviews count];
    NSMutableArray *masks = [NSMutableArray arrayWithCapacity: numSubviews];
    unsigned indexOfView = [subviews indexOfObject: aView];
    int i;
    for (i = 0; i < numSubviews; i++)
    {
        NSView *subview = [subviews objectAtIndex: i];
        unsigned mask = [subview autoresizingMask];
        [masks addObject: [NSNumber numberWithUnsignedInt: mask]];
        if (i <= indexOfView)
        {
            [subview setAutoresizingMask: NSViewMinYMargin];
        }
        else
        {
            [subview setAutoresizingMask: NSViewMaxYMargin];
        }
    }
    
    float deltaHeight;
    NSSize viewSize = [aView frame].size;
    if ([aView isHidden])
    {
        [aView setHidden: NO];
        deltaHeight = viewSize.height;
    }
    else
    {
        [aView setHidden: YES];
        deltaHeight = -viewSize.height;
    }

    [self changeWidthBy: 0.0
               heightBy: deltaHeight
                animate: YES];
    
    for (i = 0; i < numSubviews; i++)
    {
        NSView *subview = [subviews objectAtIndex: i];
        unsigned mask = [[masks objectAtIndex: i] unsignedIntValue];
        [subview setAutoresizingMask: mask];
    }
}

@end
