//
//  IsingModel.m
//  IsingCocoa
//
//  Created by Cameron Hayne on Tue May 21 2002.
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "IsingModel.h"
#import "Util.h"
#import "misc.h"

@implementation IsingModel

- (id)init
{
    self = [super init];
    if (self == nil)
        return nil;

    //NSLog(@"IsingModel.init");
    calculationLock = [[NSLock alloc] init];
    copyingLock = [[NSLock alloc] init];
    upperBoundModelWidth = [Util maxUsableWindowSize].width;
    maxUpdateRate = 1000.0; // fps
    temperature = 1.0;
    fieldIsPresent = YES;
    field = 0.0;

    init_ising();
    calc_boltz(temperature, field, boltzUp);
    initState = RandomSpins;
    
    return self;
}

- (void)dealloc
{
    //NSLog(@"IsingModel.dealloc");
    free_spins(spins);
    free_spins(newSpins);
    [calculationLock release];
    [copyingLock release];
    [super dealloc];
}

// returns a size that is acceptable to the model (as close as possible to proposedSize)
- (IsingSize)acceptableSize:(IsingSize)proposedSize
{
    // ensure that the width is a multiple of WORD_SIZE and at least 2 * WORD_SIZE
    int acceptableWidth = proposedSize.width;
    int acceptableHeight = proposedSize.height;
    int extraWidth = acceptableWidth % WORD_SIZE;
    
    // change to the closest multiple of WORD_SIZE
    if (extraWidth <= (WORD_SIZE - extraWidth)
        || (acceptableWidth / upperBoundModelWidth > 0.9))
    {
        acceptableWidth -= extraWidth;
    }
    else
    {
        acceptableWidth += (WORD_SIZE - extraWidth);
    }
    
    // sanity check
    if (acceptableWidth < WORD_SIZE)
    {
        acceptableWidth = WORD_SIZE;
    }

    IsingSize acceptableSize = makeIsingSize(acceptableWidth, acceptableHeight);
    return acceptableSize;
}

// the following method is called from the controller when the user sets the size
// Note: this method changes 'spins' and 'newSpins'
- (IsingSize)requestSize:(IsingSize)aSize modelWasReinit:(BOOL *)reinit
{
    IsingSize newSize = [self acceptableSize: aSize]; // double-check that the size is good
    if (myWidth == newSize.width && myHeight == newSize.height)
    {
        // no change in size, so we keep on going with the current calculation
        *reinit = NO;
    }
    else
    {
        myWidth  = newSize.width;
        myHeight = newSize.height;
        
        [self initialize];
        *reinit = YES;
    }

    return newSize;  // informs the sender of the model's size
}

- (void)initialize
{
    BOOL updateWasActive = updateIsActive;
    if (updateWasActive)
    {
        // stop the calculation thread
        [self stopUpdate];
    }
    
    // Allocate space for the array of spins
    free_spins(spins);
    free_spins(newSpins);
    spins    = allocate_spins(myHeight, myWidth); // # rows, # columns
    newSpins = allocate_spins(myHeight, myWidth); // # rows, # columns
    
    initialize_spins(newSpins, initState);
    copy_spins(spins, newSpins);
    
    // Reset the bookkeeping
    numSweeps = 0;
    
    if (updateWasActive)
    {
        // restart the calculation thread if it was active before
        [self startUpdate];
    }
}

- (void)setSpinConfig:(IsingInitState)desiredState
{
    BOOL updateWasActive = updateIsActive;
    if (updateWasActive)
    {
        // stop the calculation thread
        [self stopUpdate];
    }
    
    initialize_spins(newSpins, desiredState);
    copy_spins(spins, newSpins);
    
    if (updateWasActive)
    {
        // restart the calculation thread if it was active before
        [self startUpdate];
    }
}

- (BOOL)updateIsActive
{
    return updateIsActive;
}

- (void)startUpdate
{
    [calculationLock lock];  // only one calculation thread per model
    updateIsActive = YES;
    [NSThread detachNewThreadSelector: @selector(update:)
                             toTarget: self
                           withObject: nil];
}

- (void)stopUpdate
{
    updateIsActive = NO;
    // This will make the update method finish and so end the calculation thread
    // We wait for the update method to finish
    [calculationLock lock];
    [calculationLock unlock];
}

// sleeps in small intervals so as to be able to respond quickly to 'stopUpdate'
- (void)sleepUpdateThreadUntilDate:(NSDate *)endDate
{
    NSTimeInterval sleepTime = [endDate timeIntervalSinceNow];
    while (updateIsActive && sleepTime > 0.0)
    {
        NSTimeInterval delta = (sleepTime > 0.2 ? 0.2 : sleepTime);
        NSDate *endTime = [[NSDate alloc] initWithTimeIntervalSinceNow: delta];
        [NSThread sleepUntilDate: endTime];
        [endTime release];
        sleepTime -= delta;
    }
}

- (NSDate *)timeForNextUpdate
{
    // we recalculate the updatePeriod each time so that the effect of
    // changing 'maxUpdateRate' via a slider will be seen while updating.
    // Note that we avoid the use of autorelease since the pool doesn't seem
    // to get emptied while we are in the loop with no user interaction.
    NSTimeInterval updatePeriod = 1.0 / maxUpdateRate;
    NSDate *endTime = [[NSDate alloc] initWithTimeIntervalSinceNow: updatePeriod];
    return endTime; // needs to be released by caller
}

// the following is executed in the "calculation" thread
- (void)update:(id)anObject
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSDate *endTime = [self timeForNextUpdate];
    
    while (updateIsActive)
    {
        [self sleepUpdateThreadUntilDate: endTime]; // regulate the fps
        [endTime release];
        endTime = [self timeForNextUpdate];

        // the calculation works on 'newSpins'
        update_spins(newSpins, boltzUp);

        [copyingLock lock];           // ensure that 'spins' aren't in use

        // copy 'newSpins' into 'spins'
        copy_spins(spins, newSpins);
        numSweeps++;
        [copyingLock unlock];         // allow 'spins' to be used

	if (numSweeps % 5 == 0)
	{
	    calc_ising_statistics(spins,
                                  &magnetization, &nnCorrelation);
	}
    }

    [calculationLock unlock];
    [pool release];
}


// Accessors

- (float)maxUpdateRate
{
    return maxUpdateRate;
}

- (void)setMaxUpdateRate:(float)fps
{
    maxUpdateRate = fps;
}

- (IsingSize)size
{
    return makeIsingSize(myWidth, myHeight);
}

- (Spins *)lockSpinsForDisplay
{
    [copyingLock lock];
    return spins;
}

- (void)unlockSpinsForDisplay
{
    [copyingLock unlock];
}

- (int)numSweeps
{
    return numSweeps;
}

- (double)temperature
{
    return temperature;
}

- (void)setTemperature:(double)value
{
    /* avoid division by zero problems */
    if (value <= 0.0)
        value = 1.0e-9;

    temperature = value;
    calc_boltz(temperature, field, boltzUp);
}

- (BOOL)fieldIsPresent
{
    return fieldIsPresent;
}

- (void)setFieldIsPresent:(BOOL)state
{
    fieldIsPresent = state;
}

- (double)field
{
    return field;
}

- (void)setField:(double)value
{
    field = value;
    calc_boltz(temperature, field, boltzUp);
}

- (double)magnetization
{
    return magnetization;
}

- (double)nnCorrelation
{
    return nnCorrelation;
}

@end


