//
//  BlockSpinsWin.m
//  IsingCocoa
//
//  Created by Cameron Hayne on 14/04/05.
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "BlockSpinsWin.h"
#import "BlockSpinView.h"
#import "IsingValue.h"
#import "WindowSizing.h"
#import "Util.h"

// private methods
@interface BlockSpinsWin (Private)
- (void)calcBlockSpinSizes:(IsingSize)origSize;
- (void)resizeWindowForContentSize;
- (void)removeAllViews;
- (void)freeAllBlockSpins;
- (void)wantResizeNotifications:(BOOL)notificationsDesired;
@end

static int theMinBlockSpinWidth = 5;
static int theMinBlockSpinHeight = 5;
static int theGapWidth = 10;
static int theGapHeight = 10;


@implementation BlockSpinsWin

- (id)initWithSpinsFrameSize:(NSSize)spinsFrameSize
                  mainWindow:(NSWindow *)mainWindow
                   spinsSize:(IsingSize)spinsSize
                   blockSize:(IsingSize)blockSize
{
    self = [super initWithWindowNibName: @"BlockSpinsWindow"];
    if (self == nil)
        return nil;

    //NSLog(@"BlockSpinsWin: init");
    myMainWindow = mainWindow;
    myOrigSpinsSize = spinsSize;
    myBlockSize = blockSize;
    
    // turn off cascading - we want to position the window relative to the main window
    [self setShouldCascadeWindows: NO];

    // myViewFrameSize is the frame size of each of the block spin views
    myViewFrameSize = NSMakeSize(roundf(spinsFrameSize.width  / myBlockSize.width),
                                 roundf(spinsFrameSize.height / myBlockSize.height));
    if (myViewFrameSize.width < 80)
    {
        float factor = 80 / myViewFrameSize.width;
        myViewFrameSize.width  *= factor;
        myViewFrameSize.height *= factor; 
    }
    
    // create some arrays for bookkeeping
    myBlockSpinSizes = [[NSMutableArray alloc] initWithCapacity: 10];
    myViews = [[NSMutableArray alloc] initWithCapacity: 10];
    myBlockSpins = [[NSMutableArray alloc] initWithCapacity: 10];

    return self;
}

- (void)dealloc
{
    //NSLog(@"BlockSpinsWin: dealloc");
    [myBlockSpinSizes removeAllObjects];
    [self removeAllViews];
    [self freeAllBlockSpins];
    
    [myBlockSpinSizes release];
    [myViews release];
    [myBlockSpins release];
    
    [super dealloc];
}

- (void)windowDidLoad
{
    //NSLog(@"BlockSpinsWin: windowDidLoad");
    NSWindow *window = [self window];
    [window useOptimizedDrawing: YES]; // does this make a difference?
    
    [self updateIsingViewSpinsSize: myOrigSpinsSize]; // creates the sub-views
    
    // we want the blockSpin window to appear just below the main window
    NSRect mainWindowFrame = [myMainWindow frame];
    NSPoint desiredPosition = NSMakePoint(mainWindowFrame.origin.x,
                                          mainWindowFrame.origin.y - 2);
    if (desiredPosition.y - [window frame].size.height < 0)
    {
        // put the window at the top of the screen if it would be too far down
        desiredPosition.y = [window maxUsableWindowSize].height;
    }
    [window setFrameTopLeftPoint: desiredPosition];
    
    // make the window defer to the main window so that menu commands work
    [window setNextResponder: myMainWindow];

    // delay the appearance of the window until after it has finished resizing
    [window performSelector: @selector(orderFront:)
                 withObject: self
                 afterDelay: 0.0];
}

- (NSString *)windowTitleForDocumentDisplayName:(NSString *)displayName
{
    NSString *title = [NSString stringWithFormat:@"BlockSpins (%d x %d) (%@)",
                        myBlockSize.width, myBlockSize.height,
                        displayName];
    return title;
}

- (void)removeAllViews
{
    int numViews = [myViews count];
    for (int i = 0; i < numViews; i++)
    {
        BlockSpinView *view = [myViews objectAtIndex: i];
        [view removeFromSuperview];
        [view release];
    }
    
    [myViews removeAllObjects];
}

- (void)freeAllBlockSpins
{
    int numBlockSpins = [myBlockSpins count];
    for (int i = 0; i < numBlockSpins; i++)
    {
        NSValue *value = [myBlockSpins objectAtIndex: i];
        Spins *blockSpins = [value pointerValue];
        free_spins(blockSpins);
    }
    
    [myBlockSpins removeAllObjects];
}

- (void)calcBlockSpinSizes:(IsingSize)origSize
{
    [myBlockSpinSizes removeAllObjects];
    
    IsingSize blockSpinsSize = calc_block_spins_size(origSize, myBlockSize);
    while (blockSpinsSize.width > theMinBlockSpinWidth
           && blockSpinsSize.height > theMinBlockSpinHeight)
    {
        [myBlockSpinSizes addObject: [NSValue valueWithIsingSize: blockSpinsSize]];

        blockSpinsSize = calc_block_spins_size(blockSpinsSize, myBlockSize);
    }
}

#pragma mark Resizing methods

- (void)windowDidResize:(NSNotification *)notification
{
    //NSLog(@"BlockSpinsWin windowDidResize");
    NSWindow *window = [notification object];
    NSRect contentRect = [window contentRectForFrameRect: [window frame]];
    float deltaWidth = contentRect.size.width - myWindowContentSize.width;
    myWindowContentSize = contentRect.size;
    
    int numViews = [myViews count];
    float ratio = myViewFrameSize.width / myViewFrameSize.height;
    myViewFrameSize.width += (deltaWidth / numViews);
    if (myViewFrameSize.width < 80)
    {
        myViewFrameSize.width = 80;
    }
    myViewFrameSize.height = myViewFrameSize.width / ratio;
    
    // a bit brute force, but we recreate all the sub-views
    [self updateIsingViewSpinsSize: myOrigSpinsSize];
}

- (void)resizeWindowForContentSize
{
    //NSLog(@"BlockSpinsWin.resizeWindowForContentSize");
    [self wantResizeNotifications: NO];
    NSWindow *window = [self window];
    [window resizeForContentSize: myWindowContentSize
                         animate: NO];
    [self wantResizeNotifications: YES];
}

- (void)wantResizeNotifications:(BOOL)notificationsDesired
{
    NSWindow *window = [self window];
    if (notificationsDesired)
    {
        [[NSNotificationCenter defaultCenter]
            addObserver: self
               selector: @selector(windowDidResize:)
                   name: NSWindowDidResizeNotification
                 object: window];
    }
    else
    {
        [[NSNotificationCenter defaultCenter]
            removeObserver: self
                      name: NSWindowDidResizeNotification
                    object: window];
    }
}

#pragma mark Spin update methods

- (void)updateIsingViewSpinsSize:(IsingSize)origSize
{    
    [self removeAllViews];
    [self freeAllBlockSpins];

    if (origSize.width != myOrigSpinsSize.width
        || origSize.height != myOrigSpinsSize.height)
    {
        // change myViewFrameSize to accord with the new configuration
        float widthFactor  = origSize.width / ((float)myOrigSpinsSize.width);
        float heightFactor = origSize.height / ((float)myOrigSpinsSize.height);
        myViewFrameSize.width  *= (widthFactor / (widthFactor + heightFactor));
        myViewFrameSize.height *= (heightFactor / (widthFactor + heightFactor));
        if (myViewFrameSize.width < 80)
        {
            // ensure width of each BlockSpinView is at least 80 pixels
            float factor = 80 / myViewFrameSize.width;
            myViewFrameSize.width  *= factor;
            myViewFrameSize.height *= factor; 
        }
        myOrigSpinsSize = origSize;
    }

    [self calcBlockSpinSizes: origSize];

    NSWindow *window = [self window];
    float maxHeight = 0;
    NSPoint origin = NSMakePoint(2 * theGapWidth, 2 * theGapHeight);
    int numBlockSpinSizes = [myBlockSpinSizes count];
    for (int i = 0; i < numBlockSpinSizes; i++)
    {
        NSValue *value = [myBlockSpinSizes objectAtIndex: i];
        IsingSize size = [value isingSizeValue];

        NSRect frame = NSMakeRect(origin.x, origin.y,
                                  myViewFrameSize.width, myViewFrameSize.height);
        BlockSpinView *view = [[BlockSpinView alloc] initWithFrame: frame];
        
        [myViews addObject: view];
        [[window contentView] addSubview: view];
        [view updateIsingViewSpinsSize: size]; // we do this after any resizing has happened
        
        NSRect viewFrame = [view frame];
        if (viewFrame.size.height > maxHeight)
        {
            maxHeight = viewFrame.size.height;
        }
        origin.x += viewFrame.size.width;
        origin.x += theGapWidth;
    }
    
    myWindowContentSize = NSMakeSize(origin.x + theGapWidth,
                                     maxHeight + 3 * theGapHeight);
    if (myWindowContentSize.width < 150)
    {
        myWindowContentSize.width = 150;
    }

    [Util performOnce: @selector(resizeWindowForContentSize)
           withTarget: self
           withObject: nil
           afterDelay: 0.0];
}

- (void)updateIsingViewSpins:(Spins *)origSpins
{
    [self freeAllBlockSpins];

    Spins *spins = origSpins;
    int numViews = [myViews count];
    for (int i = 0; i < numViews; i++)
    {
        BlockSpinView *view = [myViews objectAtIndex: i];
        spins = calc_block_spins(spins, myBlockSize);
        [view updateIsingViewSpins: spins];
        [myBlockSpins addObject: [NSValue valueWithPointer: spins]];
    }
}

// we need this to conform to the <IsingView> protocol
- (NSWindow *)isingViewWindow
{
    return [self window];
}

@end
