//
//  IsingSpinView.m
//  IsingCocoa
//
//  Created by Cameron Hayne on Tue May 21 2002.
//  Copyright Hayne of Tintagel 2005
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "IsingSpinView.h"

@implementation IsingSpinView

// TODO: should experiment to see if using some other number of bitsPerSample
// might give a faster display. E.g. 1 (since black & white) or 16 (to match screen depth)
static NSBitmapImageRep *createImageRep(int width, int height)
{
    NSBitmapImageRep *imageRep = [[NSBitmapImageRep alloc]
                      initWithBitmapDataPlanes: 0
                                    pixelsWide: width
                                    pixelsHigh: height
                                 bitsPerSample: 8
                               samplesPerPixel: 1
                                      hasAlpha: NO
                                      isPlanar: NO
                                colorSpaceName: NSDeviceWhiteColorSpace
                                   bytesPerRow: 0
                                  bitsPerPixel: 0];
    return imageRep;
}

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame: frame];
    if (self == nil)
        return nil;
    
    //NSLog(@"IsingSpinView frame %@", NSStringFromRect(frame));
    return self;
}

- (void)dealloc
{
    //NSLog(@"IsingSpinView.dealloc");
    // do we need to free mySpins here?
    // I don't think so since it is just a pointer to data held by IsingViewCtl
    [myImageRep release];
    [super dealloc];
}

- (void)awakeFromNib
{
    NSRect bounds = [self bounds];
    mySpins = nil;
    IsingSize isingSize = makeIsingSize(bounds.size.width, bounds.size.height);
    [self updateIsingViewSpinsSize: isingSize];
}

// called from awakeFromNib and from IsingWindowCtl.initModelSize
- (void)updateIsingViewSpinsSize:(IsingSize)size
{
    NSSize boundsSize = [self bounds].size;
    if (size.width != boundsSize.width || size.height != boundsSize.height)
    {
        //NSLog(@"IsingSpinView.setSpinsSize: setting bounds size");
        mySpins = nil;
        [self setBoundsSize: NSMakeSize(size.width, size.height)];
    }
    
    // record the width & height so that we can reset the bounds to this
    // after the view gets resized
    // (we want the bounds to always be exactly the same as the spins size,
    //  thus a resize has a zooming effect on the display)
    myWidth  = size.width;
    myHeight = size.height;
    //NSLog(@"spinView: width: %d height: %d", myWidth, myHeight);

    // create an ImageRep for the IsingView
    [myImageRep release];
    myImageRep = createImageRep(myWidth, myHeight);

    // reset the bookkeeping variables
    myNumRedraws = 0;
}

- (void)updateIsingViewSpins:(Spins *)spins
{
    mySpins = spins;
    [self setNeedsDisplay: YES];
}

#pragma mark Display methods

- (BOOL)isOpaque
{
    // this allows the graphics system to draw faster
    return YES;
}

static void drawStringInRect(NSRect rect, NSString *str, int fontSize)
{
    NSDictionary *dict = [NSDictionary
                          dictionaryWithObject: [NSFont boldSystemFontOfSize: fontSize]
                                        forKey: NSFontAttributeName];
    NSAttributedString *astr = [[NSAttributedString alloc]
                                initWithString: str
                                    attributes: dict];
    NSSize strSize = [astr size];
    float scaleFactor = (rect.size.width / strSize.width) / 2.0;
    NSPoint pt = NSMakePoint(((rect.size.width / scaleFactor) - strSize.width) / 2,
                             ((rect.size.height / scaleFactor) - strSize.height) / 2);

    // clear the rect
    [[NSColor whiteColor] set];
    NSRectFill(rect);

    [[NSGraphicsContext currentContext] saveGraphicsState];
    NSAffineTransform *transform = [NSAffineTransform transform];
    [transform scaleBy: scaleFactor];
    [transform concat];
    
    // draw the string
    [astr drawAtPoint: pt];
    [astr release];
    
    [[NSGraphicsContext currentContext] restoreGraphicsState];
}

// the main drawing routine
- (void)drawRect:(NSRect)aRect
{
    // we ignore the 'aRect' argument - possible future optimization would use it
    NSRect bounds = [self bounds];    

    // we use the spins to fill in the 'bitmapData' before drawing
    BOOL imageRepInit;
    if (mySpins == NULL)
    {
        drawStringInRect(bounds, @"Initializing...", 18);
        imageRepInit = NO;
    }
    else if (mySpins->numCols != myWidth || mySpins->numRows != myHeight)
    {
        NSString *msg = [NSString stringWithFormat:
            @"Internal Error: numCols, numRows = (%d, %d) width, height = (%d, %d)",
            mySpins->numCols, mySpins->numRows, myWidth, myHeight];
        NSLog(msg);
        
        drawStringInRect(bounds, msg, 18);
        imageRepInit = NO;
    }
    else
    {
        int         numRows = mySpins->numRows;
        int         numCols = mySpins->numCols;
        SpinWord   *spinWords = mySpins->spinWords;
        int         numWcols = num_spinwords_per_row(numCols);
        
        // If 'numCols' is not a multiple of WORD_SIZE, we need to be careful
        // not to display the "extra" padding spins
        // That is the reason for the test on (col < numCols) in the inner loop
        
        SpinWord    *spinPtr = &spinWords[numWcols + 3]; // the first real spin
        const char  whiteVal = 255, blackVal = 0;
	
	// fill in the bytes of the imageRep using 'spins'
        unsigned char   *imagePtr   = [myImageRep bitmapData];
        int             bytesPerRow = [myImageRep bytesPerRow];
        
        for (int row = 0; row < numRows; row++)
        {
            for (int wcol = 0; wcol < numWcols; wcol++)
            {
		int		col = wcol;
                SpinWord	spin = *spinPtr++;
		SpinWord	mask = 1;

		int bit;
                for (bit = 0; (bit < WORD_SIZE) && (col < numCols); bit++)
                {
		    // spin up => white, spin down => black
                    imagePtr[col] = ((spin & mask) ? whiteVal : blackVal);
		    mask <<= 1;
		    col += numWcols;  // trick of multi-spin coding
                }
            }
            
            // Skip over the right and left boundary layer spins
            spinPtr += 2;
            
            // Go to the beginning of the next row in the ImageRep.
            // Here we use 'bytesPerRow' instead of 'numCols' since there
            // might be gaps at the end of the rows (for speed of drawing)
	    imagePtr += bytesPerRow;
        }

        imageRepInit = YES;
    }

    if (imageRepInit)
    {
        // We don't want anti-aliasing
        NSGraphicsContext *gc = [NSGraphicsContext currentContext];
        [gc setShouldAntialias: NO];
        [gc setImageInterpolation: NSImageInterpolationNone];

        [myImageRep draw];
        myNumRedraws++;
    }
}

- (void)viewWillStartLiveResize
{
    [super viewWillStartLiveResize];
    // could do something else here if needed
}

- (void)viewDidEndLiveResize
{
    [super viewDidEndLiveResize];
    [self setNeedsDisplay:YES];
    // could do something else here if needed
}

- (void)resetBoundsSize
{
    [self setBoundsSize: NSMakeSize(myWidth, myHeight)];
}

- (void)resizeWithOldSuperviewSize:(NSSize)oldBoundsSize
{
    //NSLog(@"IsingView.resizeWithOldSuperviewSize: %@", NSStringFromSize(oldBoundsSize));
    [super resizeWithOldSuperviewSize: oldBoundsSize];
    
    // we want the bounds to stay the same so that resizing will result in display scaling
    [self resetBoundsSize];
}

// Accessors

- (int)numRedraws
{
    return myNumRedraws;
}

// we need this to conform to the <IsingView> protocol
- (NSWindow *)isingViewWindow
{
    return [super window];
}

@end
